/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.  
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyrights
 * 
 * Portions created by or assigned to Jabber.com, Inc. are 
 * Copyright (c) 1999-2000 Jabber.com, Inc.  All Rights Reserved.  Contact
 * information for Jabber.com, Inc. is available at http://www.jabber.com/.
 *
 * Portions Copyright (c) 1998-1999 Jeremie Miller.
 * 
 * Acknowledgements
 * 
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 * 
 * --------------------------------------------------------------------------*/

#include "conference.h"

/* Generate extended presence entry */
xmlnode add_extended_presence(cnu from, cnu to, xmlnode presence, char *status, char *reason, char *actor)
{
    xmlnode tag;
    xmlnode element;
    xmlnode item;
    xmlnode output;

    taffil useraffil;
    trole userrole;
    jid user;
    cnr room;
    
    if(presence == NULL)
    {
	output = xmlnode_dup(from->presence);
    }
    else
    {
	output = xmlnode_dup(presence);
    }

    if(from == NULL)
    {
        log_debug(NAME, "[%s] ERR: Missing user variable in add_extended_presence", FZONE);
	return output;
    }

    user = from->realid;
    room = from->room;

    tag = xmlnode_insert_tag(output,"x");
    xmlnode_put_attrib(tag, "xmlns", NS_MUC_USER);

    item = xmlnode_insert_tag(tag, "item");

    if(room->visible == 1 || is_admin(room, to->realid))
    {
        xmlnode_put_attrib(item, "jid", jid_full(jid_fix(user)));
    }
	
    useraffil = affiliation_level(room, user);
    xmlnode_put_attrib(item, "affiliation", useraffil.msg);

    userrole = role_level(room, user);
    xmlnode_put_attrib(item, "role", userrole.msg);

    log_debug(NAME, "[%s] status check: status >%s<", FZONE, status);

    /* If this is a nick change, include the new nick if available */
    if(j_strcmp(status, STATUS_MUC_CREATED) == 0)
    {
        element = xmlnode_insert_tag(output, "created");
	xmlnode_put_attrib(element,"xmlns", NS_MUC_OWNER);
	room->locked = 1;
    }
    else if(status != NULL)
    {
	log_debug(NAME, "[%s] Adding to epp: status >%s<, reason >%s<", FZONE, status, reason);

	/* If this is a nick change, include the new nick if available */
	if(j_strcmp(status, STATUS_MUC_NICKCHANGE) == 0)
	    if(xmlnode_get_data(from->nick) != NULL)
	        xmlnode_put_attrib(item, "nick", xmlnode_get_data(from->nick));

	/* Add reason if available */
        if(reason != NULL)
	{
            element = xmlnode_insert_tag(item, "reason");
	    xmlnode_insert_cdata(element, reason, -1);
	}

	/* Add actor if available */
        if(actor != NULL)
	{
            element = xmlnode_insert_tag(item, "actor");
	    xmlnode_put_attrib(element, "jid", actor);
	}

	/* Add status code if available */
        element = xmlnode_insert_tag(tag, "status");
	xmlnode_put_attrib(element,"code", status);
    }

    return output;
}

/* Used to help debug contents of xhashes */
void dbg_xhash(xht h, const char *key, void *data, void *arg)
{           
    log_debug(NAME,"[%s] Key Debug: <%s>", FZONE, key);
}

/* Is the user a Service Admin? */
int is_sadmin(cni master, jid user)
{
    if(htb_get(&master->sadmin, jid_full(jid_user(jid_fix(user)))) != NULL )
        return 1;
    else
        return 0;
}

/* Is the user an owner for that room */
int is_owner(cnr room, jid user)
{
    if(room == NULL || user == NULL)
    {
        log_debug(NAME, "[%s] ERR: Missing variable in is_owner", FZONE);
	return 0;
    }

    log_debug(NAME, "[%s] Is Owner? >%s<", FZONE, jid_full(jid_fix(user)));

    /* Server admin can override */
    if(is_sadmin(room->master, user))
	    return 1;
    else if(j_strcmp(jid_full(jid_user(jid_fix(room->creator))), jid_full(jid_user(jid_fix(user)))) == 0)
	    return 1;
    else if(htb_get(&room->owner, jid_full(jid_user(jid_fix(user)))) != NULL )
	    return 1;
    else
	    return 0;

}

/* Is the user in the admin affiliation list for that room */
int is_admin(cnr room, jid user)
{
    if(room == NULL || user == NULL)
    {
        log_debug(NAME, "[%s] ERR: Missing variable in is_admin: room=%s user=%s", FZONE, room, user);
	return 0;
    }

    log_debug(NAME, "[%s] Is Admin? >%s<", FZONE, jid_full(jid_fix(user)));

    if(is_owner(room, user))
	    return 1;

    if(htb_get(&room->admin, jid_full(jid_user(jid_fix(user)))) != NULL )
	    return 1;
    else if(htb_get(&room->admin, user->server) != NULL )
	    return 1;
    else
	    return 0;
}

/* Is the user in the moderator role list for that room */
int is_moderator(cnr room, jid user)
{
    if(room == NULL || user == NULL)
    {
        log_debug(NAME, "[%s] ERR: Missing variable in is_admin: room=%s user=%s", FZONE, room, user);
	return 0;
    }

    if(is_owner(room, user))
	return 2;

    log_debug(NAME, "[%s] Is Moderator? >%s<", FZONE, jid_full(jid_fix(user)));

    if(htb_get(&room->moderator, jid_full(jid_fix(user))) != NULL )
	return 1;
    else
	return 0;
}

/* Is the user in the participant role list for that room */
int is_participant(cnr room, jid user)
{
    if(room == NULL || user == NULL)
    {
        log_debug(NAME, "[%s] ERR: Missing variable in has_voice: room=%s user=%s", FZONE, room, user);
	return 0;
    }

    /* Every non-admin has voice in a non-moderated room */
    if(room->moderated == 0) 	
	return 1;

    /* Moderator has voice intrinsic */
    if(is_moderator(room, user))
	return 1;

    /* If moderated, check the voice list */
    if(htb_get(&room->participant, jid_full(jid_fix(user))) != NULL )
	return 1;
    else
	return 0;
}

/* Is the user in the member affiliation list for that room */
int is_member(cnr room, jid user)
{
    if(room == NULL || user == NULL)
    {
        log_debug(NAME, "[%s] ERR: Missing variable in is_member: room=%s user=%s", FZONE, room, user);
	return 0;
    }

    /* Owner is automatically a member */
    if(is_owner(room, user))
	return 1;

    /* Admin is automatically a member */
    if(is_admin(room, user))
	return 1;

    if(htb_get(&room->member, jid_full(jid_user(jid_fix(user)))) != NULL )
	    return 1;
    else if(htb_get(&room->member, user->server) != NULL )
	    return 1;
    else
	    return 0;
}

/* Is the user in the outcast affiliation list for that room */
int is_outcast(cnr room, jid user)
{
    if(room == NULL || user == NULL)
    {
        log_debug(NAME, "[%s] ERR: Missing variable in is_outcast: room=%s user=%s", FZONE, room, user);
	return 0;
    }

    if(htb_get(&room->outcast, jid_full(jid_user(jid_fix(user)))) != NULL )
	    return 1;
    else if(htb_get(&room->outcast, user->server) != NULL )
	    return 1;
    else
	    return 0;
}

/* Only return 1 if visitor */
int is_visitor(cnr room, jid user)
{
    if(room == NULL || user == NULL)
    {
        log_debug(NAME, "[%s] ERR: Missing variable in is_banne: room=%s user=%s", FZONE, room, user);
	return 0;
    }

    if(is_moderator(room, user))
	    return 0;
    else if(is_participant(room, user))
	    return 0;
    else if(htb_get(&room->remote, jid_full(jid_fix(user))) != NULL )
	    return 1;
    else
	    return 0;
}

/* Is user in the room? */
int in_room(cnr room, jid user)
{
    if(room == NULL || user == NULL)
    {
        log_debug(NAME, "[%s] ERR: Missing variable in in_room: room=%s user=%s", FZONE, room, user);
	return 0;
    }

    if(htb_get(&room->remote, jid_full(jid_fix(user))) != NULL )
	    return 1;
    else
	    return 0;
}

/* Is this a legacy client? */
int is_legacy(cnu user)
{
    cnr room;

    if(user == NULL)
    {
	log_debug(NAME, "[%s] ERR: Missing variable in is_legacy", FZONE);
	return 0;
    }

    room = user->room;

    if(room->legacy)
	return 1;
    else if(user->legacy)
	return 1;
    else
	return 0;
}

/* Is user leaving the room? */
int is_leaving(cnr room, jid user)
{
    cnu target;

    if(room == NULL || user == NULL)
    {
        log_debug(NAME, "[%s] ERR: Missing variable in is_leaving: room=%s user=%s", FZONE, room, user);
	return 0;
    }

    target = htb_get(&room->remote, jid_full(jid_fix(user)));

    if(target != NULL )
    {
	if(target->leaving == 1)    
        {		
            return 1;
	}
    }

    return 0;
}

/* Check if user is already registered */
int is_registered(cni master, char *user, char *nick)
{
    xmlnode results;

    if(user == NULL || nick == NULL)
        return 0;

    results = get_data_bynick(master, nick);

    if(results != NULL)
    {
        log_debug(NAME, "[%s] Found %s in Registered Nicks - checking [%s/%s]", FZONE, nick, user, xmlnode_get_attrib(results, "jid"));

        if(j_strcmp(user, xmlnode_get_attrib(results, "jid")) != 0)
        {
	    /* User taken by someone else */
            xmlnode_free(results);
            return -1;
        }
	else
	{
	    /* Nick belongs to me */
            xmlnode_free(results);
	    return 1;
	}
    }
    else
    {
	/* Nick is free */
        xmlnode_free(results);
	return 0;
    }
}

/* Generic alert function for user/room */
xmlnode _con_send_alert(cnu user, char *text, char *subject, char *status)
{
    xmlnode msg;
    xmlnode element;
    char body[256];
    char reason[128];
    char *type = NULL;

    cnr room = user->room;
    char *room_id = jid_full(jid_fix(room->id));

    if(is_legacy(user) == 0)
    {
	return NULL;
    }

    if(status == NULL)
    {
	    snprintf(body, 256, "%s", text);
    }
    else 
    {
	if(text == NULL)
		strcpy(reason, "None given");
	else
		snprintf(reason, 128, "%s", text);

        if(j_strcmp(status, STATUS_MUC_KICKED) == 0)
	{
	    type = "normal";
	    snprintf(body, 256, "You have been kicked from the room %s. \n Reason: %s", room_id, reason);
	}

        if(j_strcmp(status, STATUS_MUC_BANNED) == 0)
	{
	    type = "normal";
	    snprintf(body, 256, "You have been kicked and outcast from the room %s. \n Reason: %s", room_id, reason);
	}

        if(j_strcmp(status, STATUS_MUC_SHOWN_JID) == 0)
	{
	    type = "groupchat";
	    snprintf(body, 256, "This room (%s) is not anonymous", room_id);
	}

        if(j_strcmp(status, STATUS_MUC_HIDDEN_JID) == 0)
	{
	    type = "groupchat";
	    snprintf(body, 256, "This room (%s) is anonymous, except for admins", room_id);
	    status = STATUS_MUC_SHOWN_JID;
	}
    }
        
    msg = jutil_msgnew(type, jid_full(jid_fix(user->realid)) , subject, body);
    xmlnode_put_attrib(msg, "from", room_id);
    
    if(status != NULL)
    {
        element = xmlnode_insert_tag(msg,"x");
        xmlnode_put_attrib(element, "xmlns", NS_MUC_USER);
        xmlnode_put_attrib(xmlnode_insert_tag(element, "status"), "code", status);

    }

    return msg;
}

/* User alaert wrapper */
void con_send_alert(cnu user, char *text, char *subject, char *status)
{
    xmlnode msg = _con_send_alert(user, text, subject, status);

    if(msg)
    {
        deliver(dpacket_new(msg), NULL);
    }
}

/* Room status/alert wrapper */
void _con_send_room_status(const char *key, void *data, void *arg)
{
    char *status = (char*)arg;
    cnu user = (cnu)data;
    xmlnode msg = _con_send_alert(user, NULL, NULL, status);

    if(msg)
    {
        deliver(dpacket_new(msg), NULL);
    }
}

/* Send status change to a room */
void con_send_room_status(cnr room, char *status)
{
    htb_walk(&room->local, _con_send_room_status, (void*)status);
}   

/* Splice \n-delimited line into array */
char *linesplit(char **entry) 
{
    char *line;
    char *point;
    char *str = "\n";

    if(!(*entry)) 
        return NULL;

    line = *entry;

    if((point = strstr(*entry, str)))
    {
        *point = 0;
        *entry = point+strlen(str);
    } 
    else
        *entry = NULL;

    return line;
}

/* Integer to String conversion */
char *itoa(int number, char *result)
{
    sprintf(result, "%d", number);
    return result;
}
		
/* Custom Debug message */
char *funcstr(char *file, char *function, int line)
{           
    static char buff[128];
    int i;
		            
    i = snprintf(buff,127,"%s:%d (%s)",file,line,function);
    buff[i] = '\0';
		            
    return buff;
}       

/* Return current date for logfile system */
int minuteget(time_t tin)
{
    time_t timef;
    char timestr[50];
    size_t timelen = 49;
    int results;

    if(tin)
	timef = tin;
    else 
        timef = time(NULL);

    strftime(timestr, timelen, "%M", localtime(&timef));
    results = j_atoi(timestr, -1);

    return results;
}

/* Return current date for logfile system */
char *timeget(time_t tin)
{
    time_t timef;
    char timestr[50];
    size_t timelen = 49;

    if(tin)
	timef = tin;
    else 
        timef = time(NULL);

    strftime(timestr, timelen, "%H:%M", localtime(&timef));

    return j_strdup(timestr);
}

/* Return current date for logfile system */
char *dateget(time_t tin)
{
    time_t timef;
    char timestr[50];
    size_t timelen = 49;

    if(tin)
	timef = tin;
    else 
        timef = time(NULL);

    strftime(timestr, timelen, "%Y-%m-%d", localtime(&timef));

    return j_strdup(timestr);
}

/* Send presence update for a user to the room */
void update_presence(cnu user)
{
    xmlnode result;
    cnr room;

    if(user == NULL)
        return;

    room = user->room;
		    
    /* Send updated presence packet */
    result = xmlnode_dup(user->presence);
    xmlnode_put_vattrib(result,"cnu",(void*)user);

    htb_walk(&room->local, con_room_sendwalk, (void*)result);
    xmlnode_free(result);
    return;
}

/* Generate custom errors for multi-item handler */
void insert_item_error(xmlnode node, char *code, char *msg)
{
    xmlnode element;

    element = xmlnode_insert_tag(node, "error");
    xmlnode_put_attrib(element, "code", code);
    xmlnode_insert_cdata(element, msg, -1);
}

/* Add user into the room roster hash */
int add_roster(cnr room, jid userid)
{
    xmlnode store;
    xmlnode node;
    char *key;

    if(room == NULL || userid == NULL)
    {
        return -1;
    }

    key = j_strdup(jid_full(jid_user(jid_fix(userid))));
    store = htb_get(&room->roster, key);

    /* User not previously registered. Set up */
    if(store == NULL)
    {
        store = xmlnode_new_tag("users");
    }
    else
    {
        node = xmlnode_get_tag(store, spools(xmlnode_pool(store), "item?jid=", jid_full(jid_fix(userid)), xmlnode_pool(store)));

        /* If already in the node, ignore */
        if(node != NULL)
        {
            free(key);
            return 0;
        }
    }

    if(userid->resource != NULL)
    {
        log_debug(NAME, "[%s] adding entry (%s) for jid (%s)", FZONE, jid_full(jid_fix(userid)), jid_full(jid_user(jid_fix(userid))));
        node = xmlnode_new_tag("item");
        xmlnode_put_attrib(node, "jid", jid_full(jid_fix(userid)));

        xmlnode_insert_node(store, node);
        xmlnode_free(node);
    }

    htb_put(&room->roster, key, store);
    free(key);

    return 1;
}

/* Remove a user from the room roster hash */
int remove_roster(cnr room, jid userid)
{
    xmlnode store;
    xmlnode node;
    char *key;

    if(room == NULL || userid == NULL)
    {
        return -1;
    }

    key = j_strdup(jid_full(jid_user(jid_fix(userid))));
    store = htb_get(&room->roster, key);

    if(store == NULL)
    {
        free(key);
        return 1;
    }

    node = xmlnode_get_tag(store, spools(xmlnode_pool(store), "item?jid=", jid_full(jid_fix(userid)), xmlnode_pool(store)));

    if(node == NULL)
    {
        free(key);
        return 1;
    }

    xmlnode_hide(node);

    node = xmlnode_get_tag(store, "item");

    if(node == NULL)
    {
        log_debug(NAME, "[%s] Removing empty entry for jid (%s)", FZONE, jid_full(jid_user(jid_fix(userid))));
	htb_zap(&room->roster, key);
        xmlnode_free(store);
    }
    else
    {
        log_debug(NAME, "[%s] Removing entry (%s) for jid (%s)", FZONE, jid_full(jid_fix(userid)), jid_full(jid_user(jid_fix(userid))));

        htb_put(&room->roster, key, store);
    }

    free(key);
    return 1;
}

/* Get the entries from the room roster hash */
xmlnode get_roster(cnr room, jid userid)
{
    xmlnode store;
    char *key;

    if(room == NULL || userid == NULL)
    {
        return NULL;
    }

    key = j_strdup(jid_full(jid_user(jid_fix(userid))));
    store = htb_get(&room->roster, key);
    free(key);

    return store;
}

char *extractAction(char *origin, pool p)
{
    int i;
    int end;
    spool sp;
    char *output;
    char in[2];

    if(origin == NULL || p == NULL)
    {
	return NULL;
    }

    sp = spool_new(p);

    end = j_strlen(origin);

    for (i = 3 ; i <= end ; i++)
    {
	in[0] = origin[i];
	in[1] = '\0';

	log_debug(NAME, "[%s] >%s< saved", FZONE, in);
	spooler(sp, in, sp);
    }
    output = spool_print(sp);

    return output;
}

/* Check Primeness for hash functions */
int isPrime(unsigned long n)
{
    int prime = 1;
    unsigned long p1,p2 , s;

    if(n > 3)
    {
        p1 = 3;
        p2 = n - 3;
        s = 9;

        while((prime = p2 % p1 ) && (s <= p2))
	{
            p1 += 2;
            p2 -= 2;
            s <<= 2;
            s++;
        }
    }
    return prime;
}

/* Used to check jids and fix case. */
jid jid_fix(jid id)
{
    unsigned char *str;

    if(id == NULL)
    {
        log_debug(NAME, "[%s] ERR - id NULL", FZONE);
        return NULL;
    }

    if(id->server == NULL || j_strlen(id->server) == 0 || j_strlen(id->server) > 255)
        return NULL;

    /* lowercase the hostname, make sure it's valid characters */
    for(str = id->server; *str != '\0'; str++)
    {
        *str = tolower(*str);
    }

    /* cut off the user */
    //if(id->user != NULL && j_strlen(id->user) > 64)
    //    id->user[64] = '\0';

    /* check for low and invalid ascii characters in the username */
    //if(id->user != NULL)
    //    for(str = id->user; *str != '\0'; str++)
    //	{
    //        *str = tolower(*str);
    //	}

    return id;
}

